"""BI Infrastructures CLI."""

import functools
from enum import Enum
from typing import Annotated

import typer
from msgspec.json import encode as json_encode
from msgspec.yaml import encode as yaml_encode
from rich.table import Table
from slices_bi_client.exceptions import ServiceException
from typer.main import console_stderr
from urllib3.connection import HTTPSConnection
from urllib3.exceptions import MaxRetryError, NewConnectionError

from . import console, err_console
from .bi_context import BIContext

infras_app = typer.Typer(no_args_is_help=True, short_help="Slices Basic Infrastructures.")


def handle_some_exceptions(func):
    """Handle specific exceptions by giving the user a useful error message instead of a confusing stacktrace."""

    @functools.wraps(func)
    def wrapper_handle_some_exceptions(*args, **kwargs):
        """Function wrapper for handle_some_exceptions."""  # noqa: D401
        try:
            return func(*args, **kwargs)
        except ServiceException as e:
            err_console.print("[red]There was a connection problem. The service might be down. :zap:")
            err_console.print(f"[red]Details: {e}")
            raise typer.Exit(code=1) from e
        except MaxRetryError as e:
            if isinstance(e.reason, NewConnectionError | HTTPSConnection):
                # example:
                #   NewConnectionError -> [Failed to establish a new connection: Errno 113] No route to host
                #   HTTPSConnection -> Failed to establish a new connection: [Errno 111] Connection refused
                err_console.print("[red]There was a connection problem. The service might be down. :zap:")
                err_console.print(f"[red]Details: {e.reason}")
                raise typer.Exit(code=1) from None
            # If it's something else, re-raise the exception and handle as usually
            raise

    return wrapper_handle_some_exceptions


class ListFormat(Enum):
    """List Formats."""

    table = "table"
    json = "json"
    yaml = "yaml"
    csv = "csv"


@infras_app.command(name="list", rich_help_panel="Commands")
@handle_some_exceptions
def list_infras(
    ctx: BIContext,
    format: Annotated[
        ListFormat,
        typer.Option("--format", "-f", help="Format in which the infrastructures should be listed."),
    ] = "table",  # type: ignore
    refresh: Annotated[
        bool,
        typer.Option("--refresh", "-r", help="Force refresh the infrastructures list", show_default=False),
    ] = False,
):
    """List Slices BI Infrastructures."""
    if refresh:
        refreshed = ctx.obj.api.reload_bi_config()
        if refreshed:
            console_stderr.print("Note: Successfully refreshed config from online source.")
        else:
            console_stderr.print(
                "Note: Config is set from the env var $SLICES_CUSTOM_CONFIG. "
                "This means the config is always fresh, and thus not refreshed again."
            )

    infras = list(ctx.obj.bi_config.bi_infrastructures.keys())

    if format == ListFormat.table:
        cols = ["ID"]

        table = Table(*cols, title="Infrastructures")

        for infra in infras:
            table.add_row(infra)

        console.print(table)

    elif format == ListFormat.json:
        console.print_json(json_encode(list(infras)).decode())
    elif format == ListFormat.csv:
        # There's only 1 column, so 1 line per infrastructure ID
        print("\n".join(infras))
    elif format == ListFormat.yaml:
        console.print("---")
        console.print(yaml_encode(list(infras)).decode())
